@@ -1,24 +1,19 @@
-'V1.15'
-'''Change Log
-new function - common
-V1.13.00
-new function - subplots
-fix text='ixs' bug in show
-show accepts 3xHxW tensors as well
-V1.12.05:
-bbs in show will accept a numpy array
-show image torch tensor bug
-'''
 __all__ = [
-    'B','Blank','BB','bbfy','C','choose','common','crop_from_bb','cv2', 'dumpdill','df2bbs','diff','find',
-    'flatten','fname','find','fname2','glob','Glob','inspect','jitter', 'L',
-    'line','loaddill','logger','extn', 'makedir', 'np', 'now','nunique','os','pd','pdfilter','parent','Path','pdb',
-    'plt','PIL','puttext','randint','rand','read','readlines','readPIL','rect','rename_batch','resize','see',
-    'set_logging_level','show','stem','stems','subplots','sys','tqdm','Tqdm','Timer','unique','uint',
+    'B','Blank','C','choose','common','crop_from_bb','diff','find',
+    'flatten','fname','find','fname2','glob','Glob','Image','inspect','jitter', 'L', 'lzip',
+    'line','lines',
+    'dumpdill','loaddill',
+    'logger','extn', 'makedir', 'np', 'now','nunique','os','pad','pd','pdfilter','parent','Path','pdb',
+    'plt','PIL','puttext','randint','rand','re','read','readPIL','rect','rename_batch','resize','rotate','see',
+    'set_logging_level','show','store_attr','stem','stems','subplots','sys','tqdm','Tqdm','trange','Timer','unique','uint','write',
+    'readlines','writelines',
+    'zip_files','unzip_file',
+    'BB','bbfy','xywh2xyXY','df2bbs',
+    'remove_duplicates','md5',
     'Info','Warn','Debug','Excep'
 ]
 
-import cv2, glob, numpy as np, pandas as pd, tqdm, os, sys
+import glob, numpy as np, pandas as pd, tqdm, os, sys, re
 import PIL
 from PIL import Image
 try:
@@ -34,6 +29,7 @@ import matplotlib.pyplot as plt
 import matplotlib.patheffects as path_effects
 import pdb, datetime, dill
 from pathlib import Path
+Path.ls = lambda x: list(x.iterdir())
 try:
     from loguru import logger
 except:
@@ -49,6 +45,11 @@ Warn  = lambda x: logger.warning(x)
 Debug = lambda x: logger.debug(x)
 Excep = lambda x: logger.exception(x)
 
+try:
+    import cv2
+    __all__ += ['cv2']
+except: logger.warning('Skipping cv2 import')
+
 import time
 class Timer:
     def __init__(self, N):
@@ -61,19 +62,28 @@ class Timer:
         elapsed = time.time() - self.start
         print('\r{}\t{}/{} ({:.2f}s - {:.2f}s remaining)'.format(info, ix+1, self.N, elapsed, (self.N-ix)*(elapsed/(ix+1))), end='')
 
-line = lambda N=66: print('='*N)
+old_line = lambda N=66: print('='*N)
+def line(string='', lw=66, upper=True, pad='='):
+    i = string.center(lw, pad)
+    if upper: i = i.upper()
+    print(i)
+
+def lines(n=3, string='', **kwargs):
+    assert n//2 == (n-1)//2, '`n` should be odd'
+    for _ in range(n//2): line(**kwargs)
+    line(string=string, **kwargs)
+    for _ in range(n//2): line(**kwargs)
+
 def see(*X, N=66): list(map(lambda x: print('='*N+'\n{}'.format(x)), X))+[print('='*N)]
 def flatten(lists): return [y for x in lists for y in  x]
 unique = lambda l: list(sorted(set(l)))
 nunique = lambda l: len(set(l))
 def choose(List, n=1):
     if n == 1: return List[randint(len(List))]
-    else:
-        _ns = np.arange(len(List))
-        np.random.shuffle(_ns)
-        return [List[ix] for ix in _ns[:n]]
+    else:      return [choose(List) for _ in range(n)]
+
 rand = lambda : ''.join(choose(list('1234567890qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM'), n=6))
-def find(item, List):
+def find(item, List, match_stem=False):
     '''Find an `item` in a `List`
     >>> find('abc', ['ijk','asdfs','dfsabcdsf','lmnop'])
     'dgsabcdsf'
@@ -81,11 +91,13 @@ def find(item, List):
     ['/tmp/file1.jpg', '/tmp/file1.png']
     '''
     filtered = [i for i in List if item in i]
+    if match_stem and len(filtered) > 1:
+        filtered = [f for f in filtered if stem(f)==item]
     if len(filtered) == 1: return filtered[0]
     return filtered
 def inspect(*arrays, **kwargs):
     '''
-    shows shape, min, max and mean of an array/list of oreys
+    shows shape, min, max and mean of an array/list/dict of oreys
     Usage:
     >>> inspect(arr1, arr2, arr3, [arr4,arr5,arr6], arr7, [arr8, arr9],...)
     where every `arr` is  assume to have a .shape, .min, .max and .mean methods
@@ -96,9 +108,10 @@ def inspect(*arrays, **kwargs):
         if ',' in names: names = names.split(',')
         assert len(names) == len(arrays), 'Give as many names as there are tensors to inspect'
     line()
+
     for ix, arr in enumerate(arrays):
         name = '\t'*depth
-        name = name + f'{names[ix].upper().strip()}:\n' if names is not None else name
+        name = name + f'{names[ix].upper().strip()}:\n' + name if names is not None else name
         name = name
         typ = type(arr).__name__
 
@@ -107,9 +120,18 @@ def inspect(*arrays, **kwargs):
                 print('[]')
             else:
                 print(f'{name}List Of {len(arr)} items')
-                inspect(*arr[:3], depth=depth+1)
-                print('\t'*(depth+1)+f'... ... {len(arr) - 3} more items')
-                line()
+                inspect(*arr[:5], depth=depth+1)
+                if len(arr) > 5:
+                    print('\t'*(depth+1)+f'... ... {len(arr) - 5} more items')
+
+        elif isinstance(arr, dict):
+            print(f'{name}Dict Of {len(arr)} items')
+            for ix,(k,v) in enumerate(arr.items()):
+                # print(f'\t'*(depth)+f' {k}'.upper())
+                inspect(v, depth=depth+1, names=[k])
+                if ix == 4: break
+            if len(arr) > 5:
+                print('\t'*(depth)+f'... ... {len(arr) - 5} more items')
 
         elif hasattr(arr, 'shape'):
             sh, m, M, dtype = arr.shape, arr.min(), arr.max(), arr.dtype
@@ -121,11 +143,13 @@ def inspect(*arrays, **kwargs):
             ln = len(arr)
             print(f'{name}{typ} Length: {ln}')
             line()
+
 randint = lambda high: np.random.randint(high)
 def Tqdm(x, total=None, desc=None):
     total = len(x) if total is None else total
     return tqdm.tqdm(x, total=total, desc=desc)
 
+from tqdm import trange
 now = lambda : str(datetime.datetime.now())[:-10].replace(' ', '_')
 
 def read(fname, mode=0):
@@ -138,13 +162,20 @@ def readPIL(fname, mode='RGB'):
     return Image.open(str(fname)).convert(mode.upper())
 
 def crop_from_bb(im, bb):
+    if isinstance(bb, list):
+        return [crop_from_bb(im, _bb) for _bb in bb]
     x,y,X,Y = bb
+    if max(x,y,X,Y) < 1.5:
+        h,w = im.shape[:2]
+        x,y,X,Y = BB(bb).absolute((h,w))
     return im.copy()[y:Y,x:X]
+
 def rect(im, bb, c=None, th=2):
-    c = (0,255,0) if c is None else c
+    c = 'g' if c is None else c
+    _d = {'r':(255,0,0), 'g':(0,255,0), 'b':(0,0,255)}
+    c = _d[c] if isinstance(c, str) else c
     x,y,X,Y = bb
     cv2.rectangle(im, (x,y), (X,Y), c, th)
-
 def B(im, th=180):
     'Binarize Image'
     return 255*(im > th).astype(np.uint8)
@@ -190,25 +221,32 @@ def rename_batch(folder, func, debug=False, one_file=False):
             os.rename(source, destin)
         # !echo {source.replace(' ','\ ')} --\> {destin.replace(' ','\ ')} >> {logfile}
         if one_file: break
+
 def common(a, b):
     """Wrapper around set intersection"""
     x = list(set(a).intersection(set(b)))
     logger.info(f'{len(x)} items found common from containers of {len(a)} and {len(b)} items respectively')
     return sorted(x)
-def diff(a, b, rev=False):
-    if not rev:
-        o = list(sorted(set(a) - set(b)))
-        logger.info(f'{len(o)} items found to differ')
-        return o
-    else:
-        o = list(sorted(set(b) - set(a)))
-        logger.info(f'{len(o)} items found to differ')
-        return o
+
+def diff(a, b, rev=False, silent=False):
+    if not rev: o = list(sorted(set(a) - set(b)))
+    else      : o = list(sorted(set(b) - set(a)))
+    if not silent: logger.info(f'{len(o)} items found to differ')
+    return o
+
 def puttext(im, string, org, scale=1, color=(255,0,0), thickness=2):
     x,y = org
     org = x, int(y+30*scale)
     cv2.putText(im, str(string), org, cv2.FONT_HERSHEY_COMPLEX, scale, color, thickness)
 
+def rotate(im, angle, return_type=np.ndarray):
+    pad = np.median(np.array(im))
+    pad = int(pad)
+    if isinstance(im, np.ndarray):
+        im = Image.fromarray(im)
+    im = im.rotate(angle, expand=1, fillcolor=(pad,pad,pad))
+    return np.array(im)
+
 def show(img=None, ax=None, title=None, sz=None, bbs=None, confs=None,
          texts=None, bb_colors=None, cmap='gray', grid=False,
          save_path=None, text_sz=15, df=None, **kwargs):
@@ -242,16 +280,6 @@ def show(img=None, ax=None, title=None, sz=None, bbs=None, confs=None,
             pass
         bbs = df2bbs(df) # assumes df has 'x,y,X,Y' columns
     if isinstance(texts, pd.core.series.Series): texts = texts.tolist()
-    if texts is not None:
-        if hasattr(texts, 'shape'):
-            if isinstance(texts, torch.Tensor): texts = texts.cpu().detach().numpy()
-            texts = texts.tolist()
-        if texts is 'ixs': texts = [i for i in range(len(bbs))]
-        if callable(texts): texts = [texts(bb) for bb in bbs]
-        assert len(texts) == len(bbs), 'Expecting as many texts as bounding boxes'
-        texts = list(map(str, texts))
-        texts = ['*' if len(t.strip())==0 else t for t in texts]
-        [puttext(ax, text.replace('$','\$'), tuple(bbs[ix][:2]), size=text_sz) for ix,text in enumerate(texts)]
     if confs:
         colors = [[255, 0, 0], [223, 111, 0], [191, 191, 0], [79, 159, 0], [0, 128, 0]]
         bb_colors = [colors[ int(cnf*5)-1 ] for cnf in confs]
@@ -267,16 +295,30 @@ def show(img=None, ax=None, title=None, sz=None, bbs=None, confs=None,
         if hasattr(bbs, 'shape'):
             if isinstance(bbs, torch.Tensor): bbs = bbs.cpu().detach().numpy()
             bbs = bbs.astype(np.uint32).tolist()
-
+        _x_ = np.array(bbs).max()
+        rel = True if _x_ < 1.5 else False
+        if rel: bbs = [BB(bb).absolute((h,w)) for bb in bbs]
         bb_colors = [[randint(255) for _ in range(3)] for _ in range(len(bbs))] if bb_colors is 'random' else bb_colors
+        bb_colors = [bb_colors]*len(bbs) if isinstance(bb_colors, str) else bb_colors
         bb_colors = [None]*len(bbs) if bb_colors is None else bb_colors
         img = C(img) if len(img.shape) == 2 else img
         [rect(img, tuple(bb), c=bb_colors[ix], th=th) for ix,bb in enumerate(bbs)]
-
+    if texts is not None:
+        if hasattr(texts, 'shape'):
+            if isinstance(texts, torch.Tensor): texts = texts.cpu().detach().numpy()
+            texts = texts.tolist()
+        if texts is 'ixs': texts = [i for i in range(len(bbs))]
+        if callable(texts): texts = [texts(bb) for bb in bbs]
+        assert len(texts) == len(bbs), 'Expecting as many texts as bounding boxes'
+        texts = list(map(str, texts))
+        texts = ['*' if len(t.strip())==0 else t for t in texts]
+        [puttext(ax, text.replace('$','\$'), tuple(bbs[ix][:2]), size=text_sz) for ix,text in enumerate(texts)]
     if title: ax.set_title(title, fontdict=kwargs.pop('fontdict', None))
     ax.imshow(img, cmap=cmap, **kwargs)
 
-    if not grid: ax.set_axis_off()
+    if grid: ax.grid()
+    else   : ax.set_axis_off()
+
     if save_path:
         fig.savefig(save_path)
         return
@@ -302,10 +344,10 @@ def loaddill(fpath):
 
 class BB:
     def __init__(self, *bb):
-        rel = False
         # assert len(bb) == 4, 'expecting a list/tuple of 4 values respectively for (x,y,X,Y)'
         if len(bb) == 4: x,y,X,Y = bb
         elif len(bb) == 1: (x,y,X,Y), = bb
+        rel = True if max(x,y,X,Y) < 1 else False
         if not rel: x,y,X,Y = map(lambda i: int(round(i)), (x,y,X,Y))
         self.bb = x,y,X,Y
         self.x, self.y, self.X, self.Y = x,y,X,Y
@@ -328,7 +370,10 @@ class BB:
         return BB(round(sf_x*self.x), round(sf_y*self.y), round(sf_x*self.X), round(sf_y*self.Y))
     def relative(self, dim:('h','w')):
         h, w = dim
-        return BB(self.x/w, self.y/h, self.X/w, self.Y/h, rel=True)
+        return BB(self.x/w, self.y/h, self.X/w, self.Y/h)
+    def absolute(self, dim:('h','w')):
+        h, w = dim
+        return BB(self.x*w, self.y*h, self.X*w, self.Y*h)
     def local_to(self, _bb):
         x,y,X,Y = self
         a,b,A,B = _bb
@@ -346,11 +391,18 @@ class BB:
 def subplots(ims, nc=5, figsize=(5,5), **kwargs):
     if len(ims) == 0: return
     titles = kwargs.pop('titles',[None]*len(ims))
+    if isinstance(titles, str):
+        if titles=='ixs': titles = [str(i) for i in range(len(ims))]
+        else: titles = titles.split(',')
+    if len(ims) <= 5 and nc==5: nc=len(ims)
     nr = (len(ims)//nc) if len(ims)%nc==0 else (1+len(ims)//nc)
     logger.info(f'plotting {len(ims)} images in a grid of {nr}x{nc} @ {figsize}')
+    figsize = kwargs.pop('sz', figsize)
+    figsize = (figsize,figsize) if isinstance(figsize, int) else figsize
     fig, axes = plt.subplots(nr,nc,figsize=figsize)
     axes = axes.flat
     fig.suptitle(kwargs.pop('suptitle',''))
+    titles = titles.split(',') if isinstance(titles, str) else titles
     for ix,(im,ax) in enumerate(zip(ims,axes)):
         show(im, ax=ax, title=titles[ix], **kwargs)
     blank = (np.eye(100) + np.eye(100)[::-1])
@@ -373,18 +425,19 @@ class L(list):
 uint = lambda im: (255*im).astype(np.uint8)
 Blank = lambda *sh: uint(np.ones(sh))
 
-def pdfilter(df, column, condition):
+def pdfilter(df, column, condition, silent=True):
     _df = df[df[column].map(condition)]
-    logger.debug(f'Filtering {len(_df)} items out of {df}')
+    if not silent: logger.debug(f'Filtering {len(_df)} items out of {len(df)}')
     return _df
+
 def pdsort(df, column, asc=True):
     df.sort_values(column, ascending=asc)
 
 def set_logging_level(level):
     logger.remove()
-    logger.add(sys.stderr, level="INFO")
+    logger.add(sys.stderr, level=level)
 
-def resize(im:np.ndarray, sz:[float,('H','W')]):
+def resize_old(im:np.ndarray, sz:[float,('H','W')]):
     h,w = im.shape[:2]
     if isinstance(sz, float):
         frac = sz
@@ -404,15 +457,137 @@ def resize(im:np.ndarray, sz:[float,('H','W')]):
             H,W = sz
     return cv2.resize(im, (W,H))
 
-def readlines(fpath):
-    with open(fpath, 'r') as f:
+def readlines(fpath, silent=False, encoding=None):
+    with open(fpath, 'r', encoding=encoding) as f:
         lines = f.read().split('\n')
         lines = [l.strip() for l in lines if l.strip()!='']
-        Info(f'loaded {len(lines)} lines')
+        if not silent: Info(f'loaded {len(lines)} lines')
         return lines
-'''108 ways to orgasm
-* Walk on soil
-* Drink clean water, eyes closed and sitting
-* When a sneeze comes, breath in instead of out
-* Feel your heart at the moment of sleep
-'''
+
+def resize(im:np.ndarray, sz:[float,('H','W'),(str,('H','W'))]):
+    '''Resize an image based on info from sz
+    *Aspect ratio is preserved
+    Examples:
+        >>> im = np.random.rand(100,100)
+        >>> _im = resize(im, 50)                    ; assert _im.shape == (50,50)
+        >>> _im = resize(im, 0.5)                   ; assert _im.shape == (50,50)   #*
+        >>> _im = resize(im, (50,200))              ; assert _im.shape == (50,200)
+        >>> _im = resize(im, (0.5,2.0))             ; assert _im.shape == (50,200)
+        >>> _im = resize(im, (0.5,200))             ; assert _im.shape == (50,200)
+
+        >>> im = np.random.rand(50,100)
+        >>> _im = resize(im, (-1, 200))             ; assert _im.shape == (100,200) #*
+        >>> _im = resize(im, (100, -1))             ; assert _im.shape == (100,200) #*
+        >>> _im = resize(im, ('at-least',(40,400))) ; assert _im.shape == (200,400) #*
+        >>> _im = resize(im, ('at-least',(400,40))) ; assert _im.shape == (400,800) #*
+        >>> _im = resize(im, ('at-most', (40,400))) ; assert _im.shape == (40,80)   #*
+        >>> _im = resize(im, ('at-most', (400,40))) ; assert _im.shape == (20,40)   #*
+   '''
+    h,w = im.shape[:2]
+    if isinstance(sz, (tuple,list)) and isinstance(sz[0], str):
+        signal, (H,W) = sz
+        assert signal in 'at-least,at-most'.split(','),\
+            'Resize type must be one of `at-least` or `at-most`'
+        if signal == 'at-least':
+            f = max(H/h,W/w)
+        if signal == 'at-most':
+            f = min(H/h,W/w)
+        H,W = [i*f for i in [h,w]]
+    elif isinstance(sz, float):
+        frac = sz
+        H,W = [i*frac for i in [h,w]]
+    elif isinstance(sz, int):
+        H,W = sz,sz
+    elif isinstance(sz, tuple):
+        H,W = sz
+        if H == -1:
+            _,W = sz
+            f = W/w
+            H = f*h
+        elif W == -1:
+            H,_ = sz
+            f = H/h
+            W = f*w
+        elif isinstance(H, float): H = H*h
+        elif isinstance(W, float): W = W*h
+    H,W = int(H),int(W)
+    return cv2.resize(im, (W,H))
+
+def pad(im, sz, pad_value=255):
+    h,w = im.shape[:2]
+    IM = np.ones(sz)*pad_value
+    IM[:h,:w] = im
+    return IM
+
+def writelines(lines, file):
+    makedir(parent(file))
+    failed = []
+    with open(file, 'w') as f:
+        for line in lines:
+            try: f.write(f'{line}\n')
+            except: failed.append(line)
+    if failed!=[]:
+        logger.info(f'Failed to write {len(failed)} lines out of {len(lines)}')
+        return failed
+
+def zip_files(list_of_files, dest):
+    import zipfile
+    Info(f'Zipping {len(list_of_files)} files to {dest}...')
+    with zipfile.ZipFile(dest, 'w') as zipMe:
+        for file in Tqdm(list_of_files):
+            zipMe.write(file, compress_type=zipfile.ZIP_DEFLATED)
+
+def unzip_file(file, dest):
+    import zipfile
+    with zipfile.ZipFile(file, 'r') as zip_ref:
+        zip_ref.extractall(dest)
+
+def md5(fname):
+    hash_md5 = hashlib.md5()
+    with open(fname, "rb") as f:
+        for chunk in iter(lambda: f.read(4096), b""):
+            hash_md5.update(chunk)
+    return hash_md5.hexdigest()
+
+def remove_duplicates(files):
+    hashes = [md5(f) for f in files]
+    df = pd.DataFrame({'f':files, 'h': hashes})
+    x = df.drop_duplicates('h')
+    y = diff(files, x.f)
+    for i in y:
+        os.rename(i, './x')
+    # !rm ./x
+    return
+
+
+def xywh2xyXY(bbs):
+    if len(bbs) == 4 and isinstance(bbs[0], int):
+        x,y,w,h = bbs
+        return BBox(x,y,x+w,y+h)
+    return [xywh2xyXY(bb) for bb in bbs]
+
+def _store_attr(self, anno, **attrs):
+    for n,v in attrs.items():
+        if n in anno: v = anno[n](v)
+        setattr(self, n, v)
+        self.__stored_args__[n] = v
+
+def store_attr(names=None, self=None, but=None, cast=False, **attrs):
+    "Store params named in comma-separated `names` from calling context into attrs in `self`"
+    fr = sys._getframe(1)
+    args = fr.f_code.co_varnames[:fr.f_code.co_argcount]
+    if self: args = ('self', *args)
+    else: self = fr.f_locals[args[0]]
+    if not hasattr(self, '__stored_args__'): self.__stored_args__ = {}
+    anno = self.__class__.__init__.__annotations__ if cast else {}
+    if attrs: return _store_attr(self, anno, **attrs)
+    ns = re.split(', *', names) if names else args[1:]
+    but = [] if not but else but
+    _store_attr(self, anno, **{n:fr.f_locals[n] for n in ns if n not in but})
+
+
+def write(image, fpath):
+    makedir(parent(fpath))
+    cv2.imwrite(fpath, image)
+
+def lzip(*x): return list(zip(*x))
